#include <assert.h>
/*
 * Copyright (C) 2002-2005  3Dlabs Inc. Ltd.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  Redistributions in binary form must reproduce the above
 *  copyright notice, this list of conditions and the following
 *  disclaimer in the documentation and/or other materials provided
 *  with the distribution.
 *
 *  Neither the name of 3Dlabs Inc. Ltd. nor the names of its
 *  contributors may be used to endorse or promote products derived
 *  from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * The source of the shaders itself are in plain text files. Files ending in
 * .vert are vertex shaders, and files ending in .frag are fragment shaders.
 */

#include <antiglut.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>

/*
 * K: FIXME
 */
#ifdef __WINDOWS__ /*[*/
#include <io.h>
#else
#include <sys/types.h>
#include <unistd.h>
#endif /*]*/

/*
 * Structures
 */

/*
 * Shader types
 */

typedef enum {
    EVertexShader,
    EFragmentShader,
} EShaderType;

/*
 * Prototypes
 */

static int printOglError(char *file, int line);

#define printOpenGLError() printOglError(__FILE__, __LINE__)

/*
 * Get the location of a uniform variable
 */
static
GLint getUniLoc(GLuint program, const GLchar *name)
{
    GLint loc;

    loc = glGetUniformLocation(program, name);

    if (loc == -1)
        printf("No such uniform named \"%s\"\n", name);

    printOpenGLError();  /* Check for OpenGL errors */
    return loc;
}

/*
 * Print out the information log for a shader object
 */
static
void printShaderInfoLog(GLuint shader)
{
    int infologLength = 0;
    int charsWritten  = 0;
    GLchar *infoLog;

    printOpenGLError();  /* Check for OpenGL errors */

    glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &infologLength);

    printOpenGLError();  /* Check for OpenGL errors */

    if (infologLength > 0)
    {
        infoLog = (GLchar *)malloc(infologLength);
        if (infoLog == NULL)
        {
            printf("ERROR: Could not allocate InfoLog buffer\n");
            exit(1);
        }
        glGetShaderInfoLog(shader, infologLength, &charsWritten, infoLog);
        printf("Shader InfoLog:\n%s\n\n", infoLog);
        free(infoLog);
    }
    printOpenGLError();  /* Check for OpenGL errors */
}

/* 
 * Print out the information log for a program object
 */
static
void printProgramInfoLog(GLuint program)
{
    int infologLength = 0;
    int charsWritten  = 0;
    GLchar *infoLog;

    printOpenGLError();  /* Check for OpenGL errors */

    glGetProgramiv(program, GL_INFO_LOG_LENGTH, &infologLength);

    printOpenGLError();  /* Check for OpenGL errors */

    if (infologLength > 0)
    {
        infoLog = (GLchar *)malloc(infologLength);
        if (infoLog == NULL)
        {
            printf("ERROR: Could not allocate InfoLog buffer\n");
            exit(1);
        }
        glGetProgramInfoLog(program, infologLength, &charsWritten, infoLog);
        printf("Program InfoLog:\n%s\n\n", infoLog);
        free(infoLog);
    }
    printOpenGLError();  /* Check for OpenGL errors */
}


static int
readShader(char *fileName, EShaderType shaderType, char **shaderText)
{
    /*
     * Reads a shader from the supplied file and returns the shader in the
     * arrays passed in. Returns 1 if successful, 0 if an error occurred.
     * The parameter size is an upper limit of the amount of bytes to read.
     * It is ok for it to be too big.
     */
    FILE *fh;
    int shader_size;
    char name[256] = "glprogs/";
    int count;

    strcat(name, fileName);

    switch (shaderType) 
    {
        case EVertexShader:
            strcat(name, ".vert");
            break;
        case EFragmentShader:
            strcat(name, ".frag");
            break;
        default:
            printf("ERROR: unknown shader file type\n");
            exit(1);
            break;
    }

    fh = fopen(name, "r");
    if (!fh)
	{
		perror("fopen");
        return -1;
	}
    /*
     * Get the shader from a file.
     */
    fseek(fh, 0, SEEK_END);
    shader_size = ftell(fh);
    fseek(fh, 0, SEEK_SET);
    if(!shader_size)
	{
		printf("Empty shader source \"%s\"\n", name);
		fclose(fh);
        return -1;
	}

    *shaderText = (GLchar *) malloc(shader_size + 1);
    count = (int) fread(*shaderText, 1, shader_size, fh);
    if(count != shader_size)
    {
    	fclose(fh);
    	free(*shaderText);
    	return 0;
    }
    (*shaderText)[count] = '\0';

    if (ferror(fh))
        count = 0;

    fclose(fh);
    return count;
}


static
int readShaderSource(char *fileName, GLchar **vertexShader, GLchar **fragmentShader)
{
    /*
     * Read the source code
     */
    if (!readShader(fileName, EVertexShader, vertexShader))
    {
        printf("Cannot read the file %s.vert\n", fileName);
        return 0;
    }

    if (!readShader(fileName, EFragmentShader, fragmentShader))
    {
        printf("Cannot read the file %s.frag\n", fileName);
        return 0;
    }

    return 1;
}


GLuint brickVS, brickFS, brickProg;   /* handles to objects */

static int installBrickShaders(const GLchar *brickVertex,
                        const GLchar *brickFragment)
{
    GLint  vertCompiled, fragCompiled;    /* status values */
    GLint  linked;

    /* Create a vertex shader object and a fragment shader object */

    brickVS = glCreateShader(GL_VERTEX_SHADER);
    brickFS = glCreateShader(GL_FRAGMENT_SHADER);

    /* Load source code strings into shaders */

    glShaderSource(brickVS, 1, &brickVertex, NULL);
    glShaderSource(brickFS, 1, &brickFragment, NULL);

    /*
     * Compile the brick vertex shader, and print out
     * the compiler log file.
     */

    glCompileShader(brickVS);
    printOpenGLError();  /* Check for OpenGL errors */
    glGetShaderiv(brickVS, GL_COMPILE_STATUS, &vertCompiled);
    printShaderInfoLog(brickVS);

    /* 
     * Compile the brick fragment shader, and print out
     * the compiler log file.
     */

    glCompileShader(brickFS);
    printOpenGLError();  /* Check for OpenGL errors */
    glGetShaderiv(brickFS, GL_COMPILE_STATUS, &fragCompiled);
    printShaderInfoLog(brickFS);

    if (!vertCompiled || !fragCompiled)
        return 0;

    /* Create a program object and attach the two compiled shaders */

    brickProg = glCreateProgram();
    glAttachShader(brickProg, brickVS);
    glAttachShader(brickProg, brickFS);

    /* Link the program object and print out the info log */

    glLinkProgram(brickProg);
    printOpenGLError();  /* Check for OpenGL errors */
    glGetProgramiv(brickProg, GL_LINK_STATUS, &linked);
    printProgramInfoLog(brickProg);

	if (!linked)
		return 0;

	glValidateProgram(brickProg);
	printProgramInfoLog(brickProg);

    /* Install program object as part of current state */

    glUseProgram(brickProg);

    /* Set up initial uniform values */

    glUniform3f(getUniLoc(brickProg, "BrickColor"), 0.8f, 0.3f, 0.1f);
    glUniform3f(getUniLoc(brickProg, "MortarColor"), 0.2f, 0.2f, 0.3f);
    glUniform2f(getUniLoc(brickProg, "BrickSize"), 0.38f, 0.10f);
    glUniform2f(getUniLoc(brickProg, "BrickPct"), 0.90f, 0.85f);
    glUniform3f(getUniLoc(brickProg, "LightPosition"), 0, 0, 4.0f);

    return 1;
}

int printOglError(char *file, int line)
{
    /* Returns 1 if an OpenGL error occurred, 0 otherwise. */
    GLenum glErr;
    int    retCode = 0;

    glErr = glGetError();
    while (glErr != GL_NO_ERROR)
    {
        printf("glError in file %s @ line %d\n", file, line);
        retCode = 1;
        glErr = glGetError();
    }
    return retCode;
}

/* ***********************************************************************
	drawing stuff
   ********************************************************************* */

static void drawCube(void)
{
    #define	size 1.0f
    #define scale 0.2f
    #define	delta 0.1f

    float A[3] = { size,  size,  size * scale + delta };
    float B[3] = { size,  size, -size * scale + delta };
    float C[3] = { size, -size, -size * scale };
    float D[3] = { size, -size,  size * scale };
    float E[3] = {-size,  size,  size * scale + delta };
    float F[3] = {-size,  size, -size * scale + delta };
    float G[3] = {-size, -size, -size * scale };
    float H[3] = {-size, -size,  size * scale };

    float I[3] = { 1.0f,  0.0f,  0.0f};
    float K[3] = {-1.0f,  0.0f,  0.0f};
    float L[3] = { 0.0f,  0.0f, -1.0f};
    float M[3] = { 0.0f,  0.0f,  1.0f};
    float N[3] = { 0.0f,  1.0f,  0.0f};
    float O[3] = { 0.0f, -1.0f,  0.0f};


    glBegin(GL_QUADS);

    glNormal3fv(I);

    glTexCoord2f(1,1);
    glVertex3fv(D);
    glTexCoord2f(0,1);
    glVertex3fv(C);
    glTexCoord2f(0,0);
    glVertex3fv(B);
    glTexCoord2f(1,0);
    glVertex3fv(A);

    glNormal3fv(K);

    glTexCoord2f(1,1);
    glVertex3fv(G);
    glTexCoord2f(0,1);
    glVertex3fv(H);
    glTexCoord2f(0,0);
    glVertex3fv(E);
    glTexCoord2f(1,0);
    glVertex3fv(F);

    glNormal3fv(L);

    glTexCoord2f(1,1);
    glVertex3fv(C);
    glTexCoord2f(0,1);
    glVertex3fv(G);
    glTexCoord2f(0,0);
    glVertex3fv(F);
    glTexCoord2f(1,0);
    glVertex3fv(B);

    glNormal3fv(M);

    glTexCoord2f(1,1);
    glVertex3fv(H);
    glTexCoord2f(0,1);
    glVertex3fv(D);
    glTexCoord2f(0,0);
    glVertex3fv(A);
    glTexCoord2f(1,0);
    glVertex3fv(E);

    glNormal3fv(N);

    glTexCoord2f(1,1);
    glVertex3fv(E);
    glTexCoord2f(0,1);
    glVertex3fv(A);
    glTexCoord2f(0,0);
    glVertex3fv(B);
    glTexCoord2f(1,0);
    glVertex3fv(F);

    glNormal3fv(O);
    
    glTexCoord2f(1,1);
    glVertex3fv(G);
    glTexCoord2f(0,1);
    glVertex3fv(C);
    glTexCoord2f(0,0);
    glVertex3fv(D);
    glTexCoord2f(1,0);
    glVertex3fv(H);
    
    glEnd();
#undef size
#undef scale
#undef delta

}

#define	REPX	4
#define	REPY	4
#define	REPZ	7

#define	TRANSX	3
#define	TRANSY	4
#define	TRANSZ	-5

static void NextClearColor(void)
{
    static int color = 0;

    switch( color++ )
    {
        case 0:  glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
             break;
        case 1:  glClearColor(0.2f, 0.2f, 0.3f, 1.0f);
             break;
        default: glClearColor(0.7f, 0.7f, 0.7f, 1.0f);
             color = 0;
             break;
    }
}


/* ***********************************************************************
	anti GLUT glue
   ********************************************************************* */

static GLfloat alpha = 0, beta = 0;
static GLfloat deltax = 0, deltay = 0;
static GLfloat fmin = 4 , fcenter = 15, fmax = 26, scale = 2.0f;

static
void display(void)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();

	glTranslatef(0,0,-fcenter);
	glScalef(scale,scale,scale); 

	glTranslatef(deltax,deltay,0);
	glRotatef(alpha,0,1,0);
	glRotatef(beta,1,0,0);

	drawCube();

	glFlush();
	agSwap();
}

static void
correct_frustum(void)
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum(-1,1,-1,1,fmin,fmax);
	glMatrixMode(GL_MODELVIEW);
}

static int
handle_event_check_break(ui_event_p in)
{
	switch(in->type)
	{
		case ui_key_event:
			switch(in->key.scan)
			{
				case sc_escape:
					return 1;
		        case sc_b:
					if(in->key.press)
			            NextClearColor();
		            break;
				case sc_wheelup:
				case sc_minus:
				case sc_kp_subtract:
					fmin *= 0.9f;
					fmax *= 0.9f;
					fcenter = (fmin + fmax) * 0.5f;
					scale *= 0.85f;
					correct_frustum();
					break;
				case sc_wheeldown:
				case sc_equal:
				case sc_kp_add:
					fmin *= 1.1f;
					fmax *= 1.1f;
					scale *= 1.15f;
					fcenter = (fmin + fmax) * 0.5f;
					correct_frustum();
					break;
				case sc_space:
				case sc_kp_begin:
					if(in->key.press)
					{
						alpha = beta = deltax = deltay = 0;
					}
					break;
				case sc_up:
					deltay -= 0.1f;
					break;
				case sc_kp_up:
					beta += 3;
					break;
				case sc_down:
					deltay += 0.1f;
					break;
				case sc_kp_down:
					beta -= 3;
					break;
				case sc_left:
					deltax += 0.1f;
					break;
				case sc_kp_left:
					alpha -= 3;
					break;
				case sc_right:
					deltax -= 0.1f;
					break;
				case sc_kp_right:
					alpha += 3;
					break;
			}
			break;
		case ui_wm_event:
			switch(in->configure.wm_event)
			{
				case ui_wm_close:
					return 1;
			}
			break;
		case ui_pointer_motion_event:
			if(in->button & UI_LBUTTON_MASK)
			{
				alpha += in->client_dx;
				beta += in->client_dy;
			}
			if(in->button & UI_RBUTTON_MASK)
			{
				deltax += 0.05f * in->client_dx / scale;
				deltay -= 0.05f * in->client_dy / scale;
			}
			break;
		default:
			break;
	}
	return 0;
}

static void agMainLoop()
{
	ui_event_t ev;
	for(;;)
	{
		display();
		uiWait();
		while(uiPending())
		{
			uiNextEvent(&ev);
			if(handle_event_check_break(&ev))
				return;
		}
	}
}

/*
 * Main
 */

int main(void)
{
	GLchar *VertexShaderSource, *FragmentShaderSource;
	agWindowConfig("3Dlabs Brick Shader: AntiGlut port", 800, 800, GL_FALSE);
	if(!agInit(AG_REQUIRED_2_0,GL_TRUE,GL_FALSE,GL_FALSE))
		return -1;

	printf("GLSL version: %s\n",glGetString(GL_SHADING_LANGUAGE_VERSION));
	correct_frustum();
	glDepthFunc(GL_LESS);
	glEnable(GL_DEPTH_TEST);
	NextClearColor();
	if(readShaderSource("brick", &VertexShaderSource, &FragmentShaderSource) &&
    		installBrickShaders(VertexShaderSource, FragmentShaderSource))
		agMainLoop();
	agShutdown();
	return 0;
}
